use logos::Logos;

#[derive(Logos, Debug, PartialEq, Eq, Clone)]
#[logos(skip r"[ \t\n\f]+")]
pub enum Token<'a> {
    // hack!
    // this isn't actually supposed to be in the language.
    // i just can't figure out how to automatically choose between a top level declaration
    // or a top level expression
    // so a declaration needs the keyword def until i can figure this out
    #[token("def")]
    Def,
    #[token("let")]
    Let,
    #[token("in")]
    In,
    #[token("mat")]
    Mat,
    #[regex("[\\d]+x[\\d]+", |lex| {
        let (x, y) = lex.slice().split_once('x').expect("shouldn't fail to split"); 
        // TODO: handle overflows etc
        (x.parse().expect("should only match valid u16s"), y.parse().expect("should only match valid u16s"))
    })]
    Dimensions((u16, u16)),
    #[regex("[\\d]+", |lex| lex.slice())]
    Int(&'a str),
    #[regex("[+-]?([\\d]+\\.[\\d]*|[\\d]*\\.[\\d]+)", |lex| lex.slice())]
    Float(&'a str),
    // TODO: more bigger better more complex string lexing
    // TODO: templating?
    #[regex(r#""([^"\\]|\\["\\bnfrt]|u[a-fA-F0-9]{4})*""#, |lex| lex.slice())]
    String(&'a str),
    #[token("+")]
    Plus,
    #[token("-")]
    Minus,
    #[token("*")]
    Mult,
    #[token("/")]
    Div,
    // TODO: figure out how to allow numbers in words?
    #[regex("[a-zA-Z_]+[a-zA-Z_\\-\\d]*", |lex| lex.slice().trim())]
    Word(&'a str),
    #[regex("\\$[a-zA-Z0-9_\\-]+", |lex| &lex.slice()[1..])]
    VarIdent(&'a str),
    #[regex("\\@[a-zA-Z0-9_\\-]+", |lex| &lex.slice()[1..])]
    InputIdent(&'a str),
    #[token(",")]
    Comma,
    #[token("|")]
    Pipe,
    #[token("@|")]
    MappingPipe,
    #[token("!|")]
    NullPipe,
    #[token("@")]
    At,
    #[token(">")]
    GreaterThan,
    #[token("=")]
    Equals,
    #[token(":")]
    Colon,
    #[token(";")]
    Semicolon,
    #[token("[")]
    BracketOpen,
    #[token("]")]
    BracketClose,
    #[token("(")]
    ParenOpen,
    #[token(")")]
    ParenClose,
    #[token("{")]
    BraceOpen,
    #[token("}")]
    BraceClose,
}

#[cfg(test)]
mod tests;
